#! /bin/bash
#
# MySQL database dump with MySQL root user - call as Linux root user!
# The database dump will be created as one MySQL dump file.
# Part of mysql-addons package. Original from Alex Hornyak (THX!).
#
# Usage: mysql_dumpdb.sh [-c|--compress] database [dump_pathname|- [dump_user [dump_group]]]

dumpparameters="--opt --complete-insert --databases --events --routines --triggers"
vetodatabases="INFORMATION_SCHEMA information_schema performance_schema"

EXIT_ERR="1"				# Return code on error
SED=$(which sed)                        # sed command call

# Need to be root.
if [ ! $UID -eq 0 ]; then echo "Run as root, please." >&2; exit $EXIT_ERR; fi

# Getting parameters.
#
# Optional compress switch (-c or --compress).
option=`echo $1 | $SED -n "s/^-\+//p"`; if [ ! -z "$option" ]; then shift; fi
if [ "$option" = "c" -o "$option" = "compress" ]; then compress=$(which gzip)
else compress="cat"; fi
# Mandatory database name.
database=$1; shift
if [ -z "$database" ]; then
    echo "Usage: $0 [-c|--compress] database [dump_pathname|- [dump_user [dump_group]]]" >&2; exit $EXIT_ERR
elif [[ ! "$database" =~ ^([[:alnum:]]|[_])*$ ]]; then
    # Humble input sanitizing.
    echo "Usage: $0 [-c|--compress] database [backup_path|- [dump_user [dump_group]]]" >&2; exit $EXIT_ERR
fi
# Optional backup file pathname, defaults to ./hostname.timestamp.database.sql
filename=${1:-`pwd`/`hostname`.`date +%Y%m%d%H%M%S`.$database.sql}; shift
if [ "$filename" = "-" ]; then
    # If '-' given as filename, we need output to STDOUT.
    filename=""
else
    # Must be an .sql extension (we don't want overwrite non-sql files).
    filename="${filename%.sql}.sql"
    # If we need a compressed output, must add a relevant extension to the filename.
    case "${compress##*/}" in
        "gzip")
        filename="${filename%.gz}.gz"
        ;;
    esac
fi
# Optional Linux user (backups owner) defaults to 'backup'.
dumpuser=${1:-"backup"}; shift
# Optional Linux owners group defaults to 'disk'.
dumpgroup=${1:-"disk"}; shift
# Wired permissions.
dumpperms="640"

# Checking components.
#
# MySQL command line call as MySQL root user.
mysql=$(which mysql); if [ -z "$mysql" -o ! -x "$mysql" ]; then echo "Missing MySQL binary." >&2; exit 1; fi
if [ ! -e "/etc/mysql/root.cnf" ]; then echo "Missing root.cnf file." >&2; exit 1; fi
mysql_root="env LANG=C $mysql --defaults-file=/etc/mysql/root.cnf"
# Database dump script call.
mysqldump=$(which mysqldump); if [ -z "$mysqldump" -o ! -x "$mysqldump" ]; then echo "Missing mysqldump binary." >&2; exit 1; fi
mysqldump_root="env LANG=C $mysqldump --defaults-file=/etc/mysql/root.cnf"

# Checking database given.
if [ -z "`$mysql_root -N --batch --execute='show databases;' | \
          $SED -n s/^$database$/X/p`" ]; then
    echo "MySQL database \"$database\" isn't found." >&2
    exit $EXIT_ERR
fi
# Silently skip MySQL schema databases.
for veto in $vetodatabases ""
do
    if [ "$database" = "$veto" ]; then exit 0; fi
done

# Doing some cleanup.
if [ ! -z "$filename" -a -f "$filename" ]; then rm "$filename"; fi
# Dumping.
if [ ! -z "$filename" ]; then
    # Dump into a file.
    $mysqldump_root $dumpparameters $database | $compress >"$filename"
    chown $dumpuser:$dumpgroup "$filename"; chmod $dumpperms "$filename"
else
    # Dump to STDOUT.
    $mysqldump_root $dumpparameters $database | $compress
fi
