#!/bin/bash
#
# MySQL Inspector - tries to detect MySQL hangups and revive dead mysqld.
# Run as root periodically, via /etc/cron.d in every minute recommended.
# Part of mysql-addons package.

BASENAME="/usr/bin/basename"            # basename command call
DATE="/bin/date"                        # date command call
GREP="/bin/grep"                        # grep command call
KILL="/bin/kill"                        # kill command call
MYSQL_INIT="/etc/init.d/mysql"          # mysql init script call
MYSQL_ROOT="env LANG=C /usr/bin/mysql --defaults-file=/etc/mysql/root.cnf"
PIDOF="/bin/pidof"                      # pidof command call
PS="/bin/ps"                            # ps command call
WC="/usr/bin/wc"                        # wc command call
WHOAMI="/usr/bin/whoami"                # whoami command call

EXIT_ERR=1                              # exit code on unspecified error
EXIT_OK=0                               # exit code on success
GRACETIME_KILLED=3                      # seconds to wait after killing procs
GRACETIME_MYSQLSTART=3                  # seconds to attempt start MySQL again
GRACETIME_QUERY=3                       # seconds to wait test querying again
GRACETIME_RERUN=3                       # seconds to rerun a failed test
GRACETIME_STOPPED=3                     # seconds to wait after stopping MySQL
LOGFILE="/var/log/mysql/mysql-inspector.log"
MYSQL_PROCESSES="/usr/bin/mysqld_safe /usr/sbin/mysqld logger"
MYSQL_STOPPED=3                         # MySQL status code if stopped.
MYSQL_TEST="show databases;"            # MySQL test query
MYSQL_TESTRESULT="^mysql$"              # MySQL test expected result line(!)
MYSQL_TOOMANY="too many connection"     # MySQL error message.

MSG_DUPLICATED="Another copy of $($BASENAME $0) is running - aborted!"
MSG_KILLING="Killing processes:"
MSG_NOANSWER="Failed test query (with exit code):"
MSG_ROOTNEED="Need to be root running $($BASENAME $0) - aborted!"
MSG_STARTFAIL_INIT="Failed to (re)start MySQL via init script (with exit code):"
MSG_STARTED_INIT="MySQL (re)started via init script:"
MSG_STOPFAIL_INIT="Failed to stop MySQL via init script (with exit code):"
MSG_STOPPED_INIT="MySQL stopped via init script:"
MSG_TOOMANY="MySQL status - too many connections message detetcted!"

#
# Functions
#

# Test MySQL availability via init script status call.
function test_init {

    # Local variables.
    local message status

    # Init script call.
    message=$($MYSQL_INIT status 2>&1); status=$?
    # Exit silently if MySQL is not running intentionally.
    if [ $status -eq $MYSQL_STOPPED ]; then return $EXIT_OK; fi
    # Check for "too many connections" message - need repair if found.
    if [ ! -z "$(echo -e "$message" | $GREP -i "$MYSQL_TOOMANY")" ]; then
        log "$MYSQL_TOOMANY"
        echo -e ${FUNCNAME[0]}; return $EXIT_ERR
    fi
    # Else test passed.
    return $EXIT_OK
}

# Test MySQL availability via test query.
function test_query {

    # Local variables.
    local counter message status

    # Try 1-3 test queries for OK status and containing expected result line.
    for counter in 1 2 3
    do
        message=$(echo -e "$MYSQL_TEST" | $MYSQL_ROOT 2>&1); status=$?
        # Break out from loop with first success by status.
        if [ "$status" -eq 0 ]; then break; fi
        sleep $GRACETIME_QUERY
    done
    # Search a single line containing MYSQL_TESTRESULT, need repair if isn't found.
    if [ -z "$(echo -e "$message" | $GREP -i -e "$MYSQL_TESTRESULT")" ]; then
        log "$MSG_NOANSWER $MYSQL_TEST ($status)"
        echo -e ${FUNCNAME[0]}; return $EXIT_ERR
    # Else test passed.
    fi
    return $EXIT_OK
}

# Try to repair dead MySQL.
function repair {

    # Local variables.
    local message status process processes

    # An extra test: when we recognize failed test, start them again with some
    # grace time. If rerun succeded, probably best aborting repair.
    case "$1" in
    "test_init" | "test_query" )
        sleep $GRACETIME_RERUN
        # Rerun, but the exit code is checked only.
        message=$($1); status=$?
        if [ $status -eq 0 ]; then
            # Exited with success - aborting repair.
            log "Recalled $1 returned with success - ${FUNCNAME[0]} aborted."
            return $EXIT_OK
        fi
        ;;
    esac
    # Seems to be need repairing.

    # Phase 1: stopping MySQL.
    # 1st attempt: trying to stop via initscript.
    message=$($MYSQL_INIT stop); status=$?
    if [ "$status" -eq 0 ]; then 
        log "$MSG_STOPPED_INIT $MYSQL_INIT"
        # Allow few seconds to stabilize.
        sleep $GRACETIME_STOPPED
    # On failure: trying to kill relevant processes.
    else
        # Logging failure.
        log "$MSG_STOPFAIL_INIT $MYSQL_INIT ($status)"
        processes=""
        # Enumerating relevant processes...
        for process in $MYSQL_PROCESSES
        do
            processes="$processes $(echo $($PIDOF -x $process))"
        done
        # ...then kill them!
        log "$MSG_KILLING $processes"
        message=$($KILL $processes 2>&1); status=$?
        # Allow few seconds to stabilize.
        sleep $GRACETIME_KILLED
        # TODO! - check result.
    fi
    # Here we suppose stopped MySQL - trying to start them.

    # Phase 2: starting MySQL - 3 attempts for OK status code.
    for counter in 1 2 3
    do
        message="$($MYSQL_INIT start 2>&1)"; status=$?
        # Break out from loop with first success by status.
        if [ "$status" -eq 0 ]; then break; fi
        # Leaving some grace time before attempting again.
        sleep $GRACETIME_MYSQLSTART
    done
    # That's our best...
    if [ $status -eq 0 ]; then
        log "$(echo -e "$message" | $GREP -i "mysql")"
        log "$MSG_STARTED_INIT $MYSQL_INIT"
    else
        log "$MSG_STARTFAIL_INIT $MYSQL_INIT ($status)"
    fi
    exit $status
}

# A simple logger function.
function log {

    local DATETIME="+%d/%b/%Y:%H:%M:%S %Z"
    if [ -f "$LOGFILE" -a -w "$LOGFILE" ]; then
        echo -e "[$($DATE "$DATETIME")] $($WHOAMI) ${FUNCNAME[1]} $*" >>$LOGFILE
        echo -e "$*" >&2
    else
        echo -e "[$($DATE "$DATETIME")] $($WHOAMI) ${FUNCNAME[1]} $*" >&2
    fi
    return
}

#
# Main program
#

# Root need to run.
if [ ! $UID -eq 0 ]; then 
    log "$MSG_ROOTNEED"; exit $EXIT_ERR
fi

# Stops if another copy of Inspector is already running.
instances=$($PS aux)
instances=$(echo -e "$instances" |$GREP -i $($BASENAME $0) | $GREP -iv "$GREP" | $WC -l)
if [ "$instances" -gt 2 ]; then
    log "$MSG_DUPLICATED"; exit $EXIT_ERR
fi

# Check for MySQL status via initscript, call repair if failed.
message=$(test_init); status=$?
if [ ! $status -eq 0 ]; then
    message=$(repair "$message"); status=$?
    exit $status
fi

# Try a MySQL query (1-3 times) for an OK status and expected result.
message=$(test_query); status=$?
if [ ! $status -eq 0 ]; then
    message=$(repair "$message"); status=$?
    exit $status
fi

# That's all folks - we trust in God...
exit $status
